@@ -1,7 +1,7 @@
 /*
  * libwebsockets web server application
  *
- * Copyright (C) 2010-2016 Andy Green <andy@warmcat.com>
+ * Copyright (C) 2010-2018 Andy Green <andy@warmcat.com>
  *
  * This file is made available under the Creative Commons CC0 1.0
  * Universal Public Domain Dedication.
@@ -36,6 +36,7 @@
 #else
 #include <io.h>
 #include "gettimeofday.h"
+#include <uv.h>
 
 int fork(void)
 {
@@ -44,7 +45,7 @@ int fork(void)
 }
 #endif
 
-#include "../lib/libwebsockets.h"
+#include <libwebsockets.h>
 
 #include <uv.h>
 
@@ -52,17 +53,20 @@ static struct lws_context *context;
 static char config_dir[128];
 static int opts = 0, do_reload = 1;
 static uv_loop_t loop;
-static uv_signal_t signal_outer;
+static uv_signal_t signal_outer[2];
 static int pids[32];
+void lwsl_emit_stderr(int level, const char *line);
 
 #define LWSWS_CONFIG_STRING_SIZE (32 * 1024)
 
 static const struct lws_extension exts[] = {
+#if !defined(LWS_WITHOUT_EXTENSIONS)
 	{
 		"permessage-deflate",
 		lws_extension_callback_pm_deflate,
 		"permessage-deflate"
 	},
+#endif
 	{ NULL, NULL, NULL /* terminator */ }
 };
 
@@ -98,7 +102,9 @@ void signal_cb(uv_signal_t *watcher, int signum)
 		break;
 	}
 	lwsl_err("Signal %d caught\n", watcher->signum);
-	lws_libuv_stop(context);
+	uv_signal_stop(watcher);
+	uv_signal_stop(&signal_outer[1]);
+	lws_context_destroy(context);
 }
 
 static int
@@ -107,6 +113,7 @@ context_creation(void)
 	int cs_len = LWSWS_CONFIG_STRING_SIZE - 1;
 	struct lws_context_creation_info info;
 	char *cs, *config_strings;
+	void *foreign_loops[1];
 
 	cs = config_strings = malloc(LWSWS_CONFIG_STRING_SIZE);
 	if (!config_strings) {
@@ -117,7 +124,7 @@ context_creation(void)
 	memset(&info, 0, sizeof(info));
 
 	info.external_baggage_free_on_destroy = config_strings;
-	info.max_http_header_pool = 16;
+	info.pt_serv_buf_size = 8192;
 	info.options = opts | LWS_SERVER_OPTION_VALIDATE_UTF8 |
 			      LWS_SERVER_OPTION_EXPLICIT_VHOSTS |
 			      LWS_SERVER_OPTION_LIBUV;
@@ -131,15 +138,16 @@ context_creation(void)
 	if (lwsws_get_config_globals(&info, config_dir, &cs, &cs_len))
 		goto init_failed;
 
+	foreign_loops[0] = &loop;
+	info.foreign_loops = foreign_loops;
+	info.pcontext = &context;
+
 	context = lws_create_context(&info);
 	if (context == NULL) {
 		lwsl_err("libwebsocket init failed\n");
 		goto init_failed;
 	}
 
-	lws_uv_sigint_cfg(context, 1, signal_cb);
-	lws_uv_initloop(context, &loop, 0);
-
 	/*
 	 * then create the vhosts... protocols are entirely coming from
 	 * plugins, so we leave it NULL
@@ -147,8 +155,7 @@ context_creation(void)
 
 	info.extensions = exts;
 
-	if (lwsws_get_config_vhosts(context, &info, config_dir,
-				    &cs, &cs_len))
+	if (lwsws_get_config_vhosts(context, &info, config_dir, &cs, &cs_len))
 		return 1;
 
 	return 0;
@@ -176,7 +183,7 @@ reload_handler(int signum)
 		fprintf(stderr, "root process receives reload\n");
 		if (!do_reload) {
 			fprintf(stderr, "passing HUP to child processes\n");
-			for (m = 0; m < ARRAY_SIZE(pids); m++)
+			for (m = 0; m < (int)LWS_ARRAY_SIZE(pids); m++)
 				if (pids[m])
 					kill(pids[m], SIGHUP);
 			sleep(1);
@@ -186,8 +193,10 @@ reload_handler(int signum)
 	case SIGINT:
 	case SIGTERM:
 	case SIGKILL:
+		fprintf(stderr, "master process waiting 2s...\n");
+		sleep(2); /* give children a chance to deal with the signal */
 		fprintf(stderr, "killing service processes\n");
-		for (m = 0; m < ARRAY_SIZE(pids); m++)
+		for (m = 0; m < (int)LWS_ARRAY_SIZE(pids); m++)
 			if (pids[m])
 				kill(pids[m], SIGTERM);
 		exit(0);
@@ -199,9 +208,10 @@ reload_handler(int signum)
 
 int main(int argc, char **argv)
 {
-	int n = 0, m, debug_level = 7;
+	int n = 0, budget = 100, debug_level = 1024 + 7;
 #ifndef _WIN32
-	int status, syslog_options = LOG_PID | LOG_PERROR;
+	int m;
+	int status;//, syslog_options = LOG_PID | LOG_PERROR;
 #endif
 
 	strcpy(config_dir, "/etc/lwsws");
@@ -214,16 +224,15 @@ int main(int argc, char **argv)
 			debug_level = atoi(optarg);
 			break;
 		case 'c':
-			strncpy(config_dir, optarg, sizeof(config_dir) - 1);
-			config_dir[sizeof(config_dir) - 1] = '\0';
+			lws_strncpy(config_dir, optarg, sizeof(config_dir));
 			break;
 		case 'h':
 			fprintf(stderr, "Usage: lwsws [-c <config dir>] "
-					"[-d <log bitfield>] [-D] [--help]\n");
+					"[-d <log bitfield>] [--help]\n");
 			exit(1);
 		}
 	}
-
+#ifndef _WIN32
 	/*
 	 * We leave our original process up permanently, because that
 	 * suits systemd.
@@ -245,9 +254,8 @@ int main(int argc, char **argv)
 				break;
 			/* old */
 			if (n > 0)
-				for (m = 0; m < ARRAY_SIZE(pids); m++)
+				for (m = 0; m < (int)LWS_ARRAY_SIZE(pids); m++)
 					if (!pids[m]) {
-						// fprintf(stderr, "added child pid %d\n", n);
 						pids[m] = n;
 						break;
 					}
@@ -257,9 +265,8 @@ int main(int argc, char **argv)
 
 		n = waitpid(-1, &status, WNOHANG);
 		if (n > 0)
-			for (m = 0; m < ARRAY_SIZE(pids); m++)
+			for (m = 0; m < (int)LWS_ARRAY_SIZE(pids); m++)
 				if (pids[m] == n) {
-					// fprintf(stderr, "reaped child pid %d\n", pids[m]);
 					pids[m] = 0;
 					break;
 				}
@@ -267,19 +274,13 @@ int main(int argc, char **argv)
 // !!! implemenation needed
 #endif
 	}
-
+#endif
 	/* child process */
 
-#ifndef _WIN32
-	/* we will only try to log things according to our debug_level */
-	setlogmask(LOG_UPTO (LOG_DEBUG));
-	openlog("lwsws", syslog_options, LOG_DAEMON);
-#endif
-
-	lws_set_log_level(debug_level, lwsl_emit_syslog);
+	lws_set_log_level(debug_level, lwsl_emit_stderr_notimestamp);
 
 	lwsl_notice("lwsws libwebsockets web server - license CC0 + LGPL2.1\n");
-	lwsl_notice("(C) Copyright 2010-2016 Andy Green <andy@warmcat.com>\n");
+	lwsl_notice("(C) Copyright 2010-2018 Andy Green <andy@warmcat.com>\n");
 
 #if (UV_VERSION_MAJOR > 0) // Travis...
 	uv_loop_init(&loop);
@@ -287,33 +288,39 @@ int main(int argc, char **argv)
 	fprintf(stderr, "Your libuv is too old!\n");
 	return 0;
 #endif
-	uv_signal_init(&loop, &signal_outer);
-	uv_signal_start(&signal_outer, signal_cb, SIGINT);
-	uv_signal_start(&signal_outer, signal_cb, SIGHUP);
+	uv_signal_init(&loop, &signal_outer[0]);
+	uv_signal_start(&signal_outer[0], signal_cb, SIGINT);
+	uv_signal_init(&loop, &signal_outer[1]);
+	uv_signal_start(&signal_outer[1], signal_cb, SIGHUP);
 
 	if (context_creation()) {
 		lwsl_err("Context creation failed\n");
 		return 1;
 	}
 
-	lws_libuv_run(context, 0);
+	lws_service(context, 0);
+
+	lwsl_err("%s: closing\n", __func__);
+
+	for (n = 0; n < 2; n++) {
+		uv_signal_stop(&signal_outer[n]);
+		uv_close((uv_handle_t *)&signal_outer[n], NULL);
+	}
 
-	uv_signal_stop(&signal_outer);
 	lws_context_destroy(context);
-
+	(void)budget;
 #if (UV_VERSION_MAJOR > 0) // Travis...
-	n = 0;
-	while (n++ < 1024 && uv_loop_close(&loop))
-		uv_run(&loop, UV_RUN_NOWAIT);
+	while ((n = uv_loop_close(&loop)) && --budget)
+		uv_run(&loop, UV_RUN_ONCE);
 #endif
 
-	lws_context_destroy2(context);
-
-	fprintf(stderr, "lwsws exited cleanly\n");
+	fprintf(stderr, "lwsws exited cleanly: %d\n", n);
 
 #ifndef _WIN32
 	closelog();
 #endif
 
+	context = NULL;
+
 	return 0;
 }
