@@ -1,7 +1,7 @@
 /*
  * libwebsockets - lws-plugin-ssh-base - sshd.c
  *
- * Copyright (C) 2017 Andy Green <andy@warmcat.com>
+ * Copyright (C) 2017 - 2018 Andy Green <andy@warmcat.com>
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public
@@ -23,6 +23,7 @@
 #include "lws-ssh.h"
 
 #include <string.h>
+#include <stdlib.h>
 
 void *sshd_zalloc(size_t s)
 {
@@ -61,7 +62,7 @@ lws_p32(uint8_t *p, uint32_t v)
 int
 lws_cstr(uint8_t **p, const char *s, uint32_t max)
 {
-	uint32_t n = strlen(s);
+	uint32_t n = (uint32_t)strlen(s);
 
 	if (n > max)
 		return 1;
@@ -96,13 +97,13 @@ explicit_bzero(void *p, size_t len)
 }
 
 int
-timingsafe_bcmp(const void *a, const void *b, uint32_t len)
+lws_timingsafe_bcmp(const void *a, const void *b, uint32_t len)
 {
 	const uint8_t *pa = a, *pb = b;
 	uint8_t sum = 0;
 
 	while (len--)
-		sum |= (*pa ^ *pb);
+		sum |= (*pa++ ^ *pb++);
 
 	return sum;
 }
@@ -113,15 +114,26 @@ write_task(struct per_session_data__sshd *pss, struct lws_ssh_channel *ch,
 {
 	pss->write_task[pss->wt_head] = task;
 	pss->write_channel[pss->wt_head] = ch;
-	pss->wt_head = (pss->wt_head + 1) & 3;
+	pss->wt_head = (pss->wt_head + 1) & 7;
 	lws_callback_on_writable(pss->wsi);
 }
 
+void
+write_task_insert(struct per_session_data__sshd *pss, struct lws_ssh_channel *ch,
+	   int task)
+{
+	pss->wt_tail = (pss->wt_tail - 1) & 7;
+	pss->write_task[pss->wt_tail] = task;
+	pss->write_channel[pss->wt_tail] = ch;
+	lws_callback_on_writable(pss->wsi);
+}
+
+
 void
 lws_pad_set_length(struct per_session_data__sshd *pss, void *start, uint8_t **p,
 		   struct lws_ssh_keys *keys)
 {
-	uint32_t len = *p - (uint8_t *)start;
+	uint32_t len = lws_ptr_diff(*p, start);
 	uint8_t padc = 4, *bs = start;
 
 	if (keys->full_length)
@@ -156,7 +168,7 @@ offer(struct per_session_data__sshd *pss, uint8_t *p, uint32_t len, int first,
 	char keyt[32];
 	uint8_t keybuf[256];
 
-	keylen = get_gen_server_key_25519(pss, keybuf, sizeof(keybuf));
+	keylen = (int)get_gen_server_key_25519(pss, keybuf, (int)sizeof(keybuf));
 	if (!keylen) {
 		lwsl_notice("get_gen_server_key failed\n");
 		return 1;
@@ -280,7 +292,7 @@ offer(struct per_session_data__sshd *pss, uint8_t *p, uint32_t len, int first,
 	*p++ = 0;
 	*p++ = 0;
 
-	len = p - op;
+	len = lws_ptr_diff(p, op);
 	if (payload_len)
 		/* starts at buf + 5 and excludes padding */
 		*payload_len = len - 5;
@@ -316,7 +328,7 @@ handle_name(struct per_session_data__sshd *pss)
 			kex->match_bitfield |= 1;
 		break;
 	case SSH_KEX_NL_SHK_ALGS:
-		len = get_gen_server_key_25519(pss, keybuf, sizeof(keybuf));
+		len = (int)get_gen_server_key_25519(pss, keybuf, (int)sizeof(keybuf));
 		if (!len)
 			break;
 		if (ed25519_key_parse(keybuf, len,
@@ -405,6 +417,8 @@ ssh_free(void *p)
 	free(p);
 }
 
+#define ssh_free_set_NULL(x) if (x) { ssh_free(x); (x) = NULL; }
+
 static void
 lws_ua_destroy(struct per_session_data__sshd *pss)
 {
@@ -519,20 +533,26 @@ ssh_destroy_channel(struct per_session_data__sshd *pss,
 	lwsl_notice("Failed to delete ch\n");
 }
 
+static void
+lws_ssh_exec_finish(void *finish_handle, int retcode)
+{
+	struct lws_ssh_channel *ch = (struct lws_ssh_channel *)finish_handle;
+	struct per_session_data__sshd *pss = ch->pss;
+
+	ch->retcode = retcode;
+	write_task(pss, ch, SSH_WT_EXIT_STATUS);
+	ch->scheduled_close = 1;
+	write_task(pss, ch, SSH_WT_CH_CLOSE);
+}
 
 static int
 lws_ssh_parse_plaintext(struct per_session_data__sshd *pss, uint8_t *p, size_t len)
 {
-#if defined(LWS_WITH_MBEDTLS)
-	mbedtls_rsa_context *ctx;
-#else
-	BIGNUM *bn_e = NULL, *bn_n = NULL;
-	RSA *rsa = NULL;
-#endif
+	struct lws_genrsa_elements el;
+	struct lws_genrsa_ctx ctx;
 	struct lws_ssh_channel *ch;
 	struct lws_subprotocol_scp *scp;
 	uint8_t *pp, *ps, hash[64], *otmp;
-	size_t olen;
 	uint32_t m;
 	int n;
 
@@ -638,6 +658,10 @@ again:
 					pss->parser_state = SSH_KEX_STATE_SKIP;
 					break;
 				}
+				if (!pss->kex) {
+					lwsl_notice("%s: SSH_MSG_KEXINIT: NULL pss->kex\n", __func__);
+					goto bail;
+				}
 				pss->parser_state = SSH_KEX_STATE_COOKIE;
 				pss->kex->I_C_payload_len = 0;
 				pss->kex->I_C_alloc_len = pss->msg_len;
@@ -1044,14 +1068,12 @@ again:
 			}
 
 			pss->seen_auth_req_before = 1;
-			strncpy(pss->last_auth_req_username, pss->ua->username,
-				sizeof(pss->last_auth_req_username) - 1);
-			pss->last_auth_req_username[
-			        sizeof(pss->last_auth_req_username) - 1] = '\0';
-			strncpy(pss->last_auth_req_service, pss->ua->service,
-				sizeof(pss->last_auth_req_service) - 1);
-			pss->last_auth_req_service[
-			        sizeof(pss->last_auth_req_service) - 1] = '\0';
+			lws_strncpy(pss->last_auth_req_username,
+				    pss->ua->username,
+				    sizeof(pss->last_auth_req_username));
+			lws_strncpy(pss->last_auth_req_service,
+				    pss->ua->service,
+				    sizeof(pss->last_auth_req_service));
 
 			if (strcmp(pss->ua->service, "ssh-connection"))
 				goto ua_fail;
@@ -1092,6 +1114,7 @@ again:
 
 		case SSHS_NVC_DO_UAR_PUBKEY_BLOB:
 			pss->ua->pubkey = pss->last_alloc;
+			pss->last_alloc = NULL;
 			pss->ua->pubkey_len = pss->npos;
 			/*
 			 * RFC4253
@@ -1149,6 +1172,7 @@ again:
 			}
 			lwsl_info("SSHS_DO_UAR_SIG\n");
 			pss->ua->sig = pss->last_alloc;
+			pss->last_alloc = NULL;
 			pss->ua->sig_len = pss->npos;
 			pss->parser_state = SSHS_MSG_EAT_PADDING;
 
@@ -1178,12 +1202,12 @@ again:
 			 */
 			n = 4 + 32 +
 			    1 +
-			    4 + strlen(pss->ua->username) +
-			    4 + strlen(pss->ua->service) +
+			    4 + (int)strlen(pss->ua->username) +
+			    4 + (int)strlen(pss->ua->service) +
 			    4 + 9 +
 			    1 +
-			    4 + strlen(pss->ua->alg) +
-			    4 + pss->ua->pubkey_len;
+			    4 + (int)strlen(pss->ua->alg) +
+			    4 + (int)pss->ua->pubkey_len;
 
 			ps = sshd_zalloc(n);
 			if (!ps) {
@@ -1222,76 +1246,22 @@ again:
 			 * Prepare the RSA decryption context: load in
 			 * the E and N factors
 			 */
-#if defined(LWS_WITH_MBEDTLS)
-			ctx = sshd_zalloc(sizeof(*ctx));
-			if (!ctx)
-				goto ua_fail;
-			mbedtls_rsa_init(ctx, MBEDTLS_RSA_PKCS_V15, 0);
 
+			memset(&el, 0, sizeof(el));
 			pp = pss->ua->pubkey;
 			m = lws_g32(&pp);
 			pp += m;
 			m = lws_g32(&pp);
-			if (mbedtls_mpi_read_binary(&ctx->E, pp, m))
-				lwsl_notice("mpi load E failed\n");
+			el.e[JWK_KEY_E].buf = pp;
+			el.e[JWK_KEY_E].len = m;
 			pp += m;
 			m = lws_g32(&pp);
+			el.e[JWK_KEY_N].buf = pp;
+			el.e[JWK_KEY_N].len = m;
 
-			m = mbedtls_mpi_read_binary(&ctx->N, pp, m);
-			if (m)
-				lwsl_notice("mpi load N failed %d\n", m);
-#else
-			/* Step 1:
-			 *
-			 * convert the MPI for e and n to OpenSSL BIGNUMs
-			 */
-
-			pp = pss->ua->pubkey;
-			m = lws_g32(&pp);
-			pp += m;
-			m = lws_g32(&pp);
-			pp -= 4;
-			bn_e = BN_mpi2bn(pp, m + 4, NULL);
-			if (!bn_e) {
-				lwsl_notice("mpi load E failed\n");
+			if (lws_genrsa_create(&ctx, &el))
 				goto ua_fail;
-			}
-			pp += m + 4;
-			m = lws_g32(&pp);
-			pp -= 4;
-			bn_n = BN_mpi2bn(pp, m + 4, NULL);
-			if (!bn_n) {
-				lwsl_notice("mpi load N failed %d\n", m);
-				BN_free(bn_e);
-				bn_e = NULL;
-				goto ua_fail;
-			}
 
-			/* Step 2:
-			 *
-			 * assemble the OpenSSL RSA from the BIGNUMs
-			 */
-
-			rsa = RSA_new();
-			if (!rsa) {
-				lwsl_notice("Failed to create RSA\n");
-				BN_free(bn_e);
-				BN_free(bn_n);
-				goto ua_fail;
-			}
-#if defined(LWS_HAVE_RSA_SET0_KEY)
-			if (RSA_set0_key(rsa, bn_n, bn_e, NULL) != 1) {
-				lwsl_notice("RSA_set0_key failed\n");
-				BN_free(bn_e);
-				BN_free(bn_n);
-				goto ua_fail;
-			}
-
-#else
-			rsa->e = bn_e;
-			rsa->n = bn_n;
-#endif
-#endif
 			/*
 			 * point to the encrypted signature payload we
 			 * were sent
@@ -1300,38 +1270,21 @@ again:
 			m = lws_g32(&pp);
 			pp += m;
 			m = lws_g32(&pp);
-#if defined(LWS_WITH_MBEDTLS)
-			ctx->len = m;
-#endif
 
 			/*
-			 * decrypt it, resulting in an error or some
-			 * ASN1 including the decrypted signature
+			 * decrypt it, resulting in an error, or some ASN1
+			 * including the decrypted signature
 			 */
-
 			otmp = sshd_zalloc(m);
 			if (!otmp)
 				/* ua_fail1 frees bn_e, bn_n and rsa */
 				goto ua_fail1;
-#if defined(LWS_WITH_MBEDTLS)
-			m = mbedtls_rsa_rsaes_pkcs1_v15_decrypt(ctx,
-				NULL, NULL, MBEDTLS_RSA_PUBLIC,
-				&olen, pp, otmp, m);
-#else
-			olen = 0;
-			m = RSA_public_decrypt((int)m, pp, otmp, rsa,
-					       RSA_PKCS1_PADDING);
-			if (m != (uint32_t)-1) {
-				olen = m;
-				m = 0;
-			}
-			/* the bignums are also freed by freeing the RSA */
-			RSA_free(rsa);
-#endif
-			if (!m) {
+
+			n = lws_genrsa_public_decrypt(&ctx, pp, m, otmp, m);
+			if (n > 0) {
 				/* the decrypted sig is in ASN1 format */
 				m = 0;
-				while (m < olen) {
+				while ((int)m < n) {
 					/* sig payload */
 					if (otmp[m] == 0x04 &&
 					    otmp[m + 1] == lws_genhash_size(
@@ -1348,17 +1301,15 @@ again:
 					/* otherwise skip payloads */
 					m += otmp[m + 1] + 2;
 				}
-			} else
-				lwsl_notice("decrypt failed\n");
-#if defined(LWS_WITH_MBEDTLS)
-			mbedtls_rsa_free(ctx);
-			free(ctx);
-#endif
+			}
+
 			free(otmp);
+			lws_genrsa_destroy(&ctx);
+
 			/*
 			 * if no good, m is nonzero and inform peer
 			 */
-			if (m) {
+			if (n <= 0) {
 				lwsl_notice("hash sig verify fail: %d\n", m);
 				goto ua_fail;
 			}
@@ -1412,7 +1363,7 @@ again:
 				pss->vhd->ops->disconnect_reason(
 					pss->disconnect_reason,
 					pss->disconnect_desc, pss->name);
-			ssh_free(pss->last_alloc);
+			ssh_free_set_NULL(pss->last_alloc);
 			break;
 
 			/*
@@ -1432,6 +1383,7 @@ again:
 				return -1;
 
 			pss->ch_temp->type = SSH_CH_TYPE_SESSION;
+			pss->ch_temp->pss = pss;
 			state_get_u32(pss, SSHS_NVC_CHOPEN_SENDER_CH);
 			break;
 
@@ -1498,9 +1450,11 @@ again:
 				pss->channel_doing_spawn = pss->ch_temp->server_ch;
 				if (pss->vhd->ops && pss->vhd->ops->shell &&
 				    !pss->vhd->ops->shell(pss->ch_temp->priv,
-						          pss->wsi)) {
+						          pss->wsi,
+						 lws_ssh_exec_finish, pss->ch_temp)) {
+
 					if (pss->rq_want_reply)
-						write_task(pss, pss->ch_temp,
+						write_task_insert(pss, pss->ch_temp,
 							   SSH_WT_CHRQ_SUCC);
 					pss->parser_state = SSHS_MSG_EAT_PADDING;
 					break;
@@ -1543,7 +1497,7 @@ again:
 		/* CHRQ pty-req */
 
 		case SSHS_NVC_CHRQ_TERM:
-			strncpy(pss->args.pty.term, pss->name,
+			memcpy(pss->args.pty.term, pss->name,
 				sizeof(pss->args.pty.term) - 1);
 			state_get_u32(pss, SSHS_NVC_CHRQ_TW);
 			break;
@@ -1571,7 +1525,7 @@ again:
 			if (pss->vhd->ops && pss->vhd->ops->pty_req)
 				n = pss->vhd->ops->pty_req(pss->ch_temp->priv,
 							&pss->args.pty);
-			ssh_free(pss->last_alloc);
+			ssh_free_set_NULL(pss->last_alloc);
 			if (n)
 				goto chrq_fail;
 			if (pss->rq_want_reply)
@@ -1617,10 +1571,11 @@ again:
 
 			if (pss->vhd->ops && pss->vhd->ops->exec &&
 			    !pss->vhd->ops->exec(pss->ch_temp->priv, pss->wsi,
-					    	 (const char *)pss->last_alloc)) {
-				ssh_free(pss->last_alloc);
+					    	 (const char *)pss->last_alloc,
+						 lws_ssh_exec_finish, pss->ch_temp)) {
+				ssh_free_set_NULL(pss->last_alloc);
 				if (pss->rq_want_reply)
-					write_task(pss, pss->ch_temp,
+					write_task_insert(pss, pss->ch_temp,
 						   SSH_WT_CHRQ_SUCC);
 
 				pss->parser_state = SSHS_MSG_EAT_PADDING;
@@ -1641,7 +1596,7 @@ again:
 				/* disallow it */
 				n = 0;
 
-			ssh_free(pss->last_alloc);
+			ssh_free_set_NULL(pss->last_alloc);
 			if (!n)
 				goto chrq_fail;
 
@@ -1675,7 +1630,7 @@ again:
 				n = 1;
 			}
 #endif
-			ssh_free(pss->last_alloc);
+			ssh_free_set_NULL(pss->last_alloc);
 //			if (!n)
 				goto ch_fail;
 #if 0
@@ -1766,12 +1721,12 @@ again:
 				break;
 			}
 			if (pss->parser_state == SSHS_NVC_CD_DATA_ALLOC)
-				ssh_free(pss->last_alloc);
+				ssh_free_set_NULL(pss->last_alloc);
 
 			if (ch->peer_window_est < 32768) {
 				write_task(pss, ch, SSH_WT_WINDOW_ADJUST);
 				ch->peer_window_est += 32768;
-				lwsl_notice("extra peer WINDOW_ADJUST (~ %d)\n",
+				lwsl_info("extra peer WINDOW_ADJUST (~ %d)\n",
 					    ch->peer_window_est);
 			}
 
@@ -1810,11 +1765,14 @@ again:
 			 */
 			lwsl_notice("SSH_MSG_CHANNEL_EOF: %d\n", pss->ch_recip);
 			ch = ssh_get_server_ch(pss, pss->ch_recip);
-			if (!ch)
+			if (!ch) {
+				lwsl_notice("unknown ch %d\n", pss->ch_recip);
 				return -1;
+			}
 
-			if (!ch->had_eof) {
-				ch->had_eof = 1;
+			if (!ch->scheduled_close) {
+				lwsl_notice("scheduling CLOSE\n");
+				ch->scheduled_close = 1;
 				write_task(pss, ch, SSH_WT_CH_CLOSE);
 			}
 			pss->parser_state = SSHS_MSG_EAT_PADDING;
@@ -1854,6 +1812,7 @@ again:
 			}
 
 			ch->received_close = 1;
+			ch->scheduled_close = 1;
 			write_task(pss, ch, SSH_WT_CH_CLOSE);
 			break;
 
@@ -1876,13 +1835,7 @@ ch_fail:
 			break;
 
 ua_fail1:
-#if defined(LWS_WITH_MBEDTLS)
-			mbedtls_rsa_free(ctx);
-			free(ctx);
-#else
-			/* also frees the bignums */
-			RSA_free(rsa);
-#endif
+			lws_genrsa_destroy(&ctx);
 ua_fail:
 			write_task(pss, NULL, SSH_WT_UA_FAILURE);
 ua_fail_silently:
@@ -1937,7 +1890,7 @@ parse(struct per_session_data__sshd *pss, uint8_t *p, size_t len)
 			uint8_t pt[2048];
 
 			len++;
-			cp = len;
+			cp = (uint32_t)len;
 
 			if (cp > l - pss->pa_pos)
 				cp = l - pss->pa_pos;
@@ -2003,7 +1956,7 @@ pad_and_encrypt(uint8_t *dest, void *ps, uint8_t *pp,
 
 	if (!skip_pad)
 		lws_pad_set_length(pss, ps, &pp, &pss->active_keys_stc);
-	n = pp - (uint8_t *)ps;
+	n = lws_ptr_diff(pp, ps);
 
 	if (!pss->active_keys_stc.valid) {
 		memcpy(dest, ps, n);
@@ -2024,7 +1977,7 @@ lws_callback_raw_sshd(struct lws *wsi, enum lws_callback_reasons reason,
 	struct per_session_data__sshd *pss =
 			(struct per_session_data__sshd *)user, **p;
 	struct per_vhost_data__sshd *vhd = NULL;
-	uint8_t buf[LWS_PRE + 1024], *pp, *ps = &buf[LWS_PRE + 512];
+	uint8_t buf[LWS_PRE + 1024], *pp, *ps = &buf[LWS_PRE + 512], *ps1 = NULL;
 	const struct lws_protocol_vhost_options *pvo;
 	const struct lws_protocols *prot;
 	struct lws_ssh_channel *ch;
@@ -2045,14 +1998,12 @@ lws_callback_raw_sshd(struct lws *wsi, enum lws_callback_reasons reason,
 			lws_protocol_vh_priv_get(lws_get_vhost(wsi),
 				pss->vhd->protocol);
 	else
-		vhd = (struct per_vhost_data__sshd *)
+		if (lws_get_vhost(wsi))
+			vhd = (struct per_vhost_data__sshd *)
 				lws_protocol_vh_priv_get(lws_get_vhost(wsi),
 				lws_vhost_name_to_protocol(
 					lws_get_vhost(wsi), "lws-ssh-base"));
 
-	if (!vhd && reason != LWS_CALLBACK_PROTOCOL_INIT)
-		return -1;
-
 	switch ((int)reason) {
 	case LWS_CALLBACK_PROTOCOL_INIT:
 		vhd = lws_protocol_vh_priv_zalloc(lws_get_vhost(wsi),
@@ -2096,7 +2047,7 @@ lws_callback_raw_sshd(struct lws *wsi, enum lws_callback_reasons reason,
 		 * The user code ops api_version has to be current
 		 */
 		if (vhd->ops->api_version != LWS_SSH_OPS_VERSION) {
-			lwsl_err("FATAL ops is api_version v%d but code is v%d",
+			lwsl_err("FATAL ops is api_version v%d but code is v%d\n",
 				vhd->ops->api_version, LWS_SSH_OPS_VERSION);
 			return 1;
 		}
@@ -2104,6 +2055,8 @@ lws_callback_raw_sshd(struct lws *wsi, enum lws_callback_reasons reason,
 
         case LWS_CALLBACK_RAW_ADOPT:
 		lwsl_info("LWS_CALLBACK_RAW_ADOPT\n");
+		if (!vhd)
+			return -1;
 		pss->next = vhd->live_pss_list;
 		vhd->live_pss_list = pss;
 		pss->parser_state = SSH_INITIALIZE_TRANSIENT;
@@ -2129,10 +2082,14 @@ lws_callback_raw_sshd(struct lws *wsi, enum lws_callback_reasons reason,
                 break;
 
 	case LWS_CALLBACK_RAW_CLOSE:
+		if (!pss)
+			return -1;
 		lwsl_info("LWS_CALLBACK_RAW_CLOSE\n");
 		lws_kex_destroy(pss);
 		lws_ua_destroy(pss);
 
+		ssh_free_set_NULL(pss->last_alloc);
+
 		while (pss->ch_list)
 			ssh_destroy_channel(pss, pss->ch_list);
 
@@ -2151,11 +2108,15 @@ lws_callback_raw_sshd(struct lws *wsi, enum lws_callback_reasons reason,
 		break;
 
 	case LWS_CALLBACK_RAW_RX:
+		if (!pss)
+			return -1;
 		if (parse(pss, in, len))
 			return -1;
 		break;
 
 	case LWS_CALLBACK_RAW_WRITEABLE:
+		if (!pss)
+			break;
 		n = 0;
 		o = pss->write_task[pss->wt_tail];
 		ch = pss->write_channel[pss->wt_tail];
@@ -2250,9 +2211,9 @@ lws_callback_raw_sshd(struct lws *wsi, enum lws_callback_reasons reason,
 			pp = ps + 5;
 			*pp++ = SSH_MSG_USERAUTH_BANNER;
 			if (pss->vhd && pss->vhd->ops->banner)
-				n = pss->vhd->ops->banner((char *)&buf[650],
+				n = (int)pss->vhd->ops->banner((char *)&buf[650],
 							  150 - 1,
-							  lang, sizeof(lang));
+							  lang, (int)sizeof(lang));
 			lws_p32(pp, n);
 			pp += 4;
 			strcpy((char *)pp, (char *)&buf[650]);
@@ -2272,17 +2233,18 @@ lws_callback_raw_sshd(struct lws *wsi, enum lws_callback_reasons reason,
 			 *    string    public key blob from the request
       			 */
 			n = 74 + pss->ua->pubkey_len;
-			if (n > sizeof(buf) - LWS_PRE) {
+			if (n > (int)sizeof(buf) - LWS_PRE) {
 				lwsl_notice("pubkey too large\n");
 				goto bail;
 			}
-			ps = sshd_zalloc(n);
-			if (!ps)
+			ps1 = sshd_zalloc(n);
+			if (!ps1)
 				goto bail;
-			pp = ps + 5;
+			ps = ps1;
+			pp = ps1 + 5;
 			*pp++ = SSH_MSG_USERAUTH_PK_OK;
 			if (lws_cstr(&pp, pss->ua->alg, 64)) {
-				free(ps);
+				free(ps1);
 				goto bail;
 			}
 			lws_p32(pp, pss->ua->pubkey_len);
@@ -2391,6 +2353,21 @@ lws_callback_raw_sshd(struct lws *wsi, enum lws_callback_reasons reason,
 			lwsl_info("send SSH_MSG_CHANNEL_WINDOW_ADJUST\n");
 			goto pac;
 
+		case SSH_WT_EXIT_STATUS:
+			pp = ps + 5;
+			*pp++ = SSH_MSG_CHANNEL_REQUEST;
+			lws_p32(pp, ch->sender_ch);
+			pp += 4;
+			lws_p32(pp, 11);
+			pp += 4;
+			strcpy((char *)pp, "exit-status");
+			pp += 11;
+			*pp++ = 0;
+			lws_p32(pp, ch->retcode);
+			pp += 4;
+			lwsl_info("send SSH_MSG_CHANNEL_EXIT_STATUS\n");
+			goto pac;
+
 		case SSH_WT_NONE:
 		default:
 			/* sending payload */
@@ -2441,12 +2418,12 @@ lws_callback_raw_sshd(struct lws *wsi, enum lws_callback_reasons reason,
 			pp += pss->vhd->ops->tx(ch->priv, n, pp,
 						&buf[sizeof(buf) - 1] - pp);
 
-			lws_p32(ps + m - 4, pp - (ps + m));
+			lws_p32(ps + m - 4, lws_ptr_diff(pp, (ps + m)));
 
 			if (pss->vhd->ops->tx_waiting(ch->priv) > 0)
 				lws_callback_on_writable(wsi);
 
-			ch->window -= (pp - ps) - m;
+			ch->window -= lws_ptr_diff(pp, ps) - m;
 			//lwsl_debug("our send window: %d\n", ch->window);
 
 			/* fallthru */
@@ -2479,7 +2456,7 @@ bail:
 					lws_kex_destroy(pss);
 				break;
 			case SSH_WT_UA_PK_OK:
-				free(ps);
+				free(ps1);
 				break;
 			case SSH_WT_CH_CLOSE:
 				if (ch->received_close) {
@@ -2505,10 +2482,10 @@ bail:
 
 			if (o != SSH_WT_NONE)
 				pss->wt_tail =
-					(pss->wt_tail + 1) & 3;
+					(pss->wt_tail + 1) & 7;
 		} else
 			if (o == SSH_WT_UA_PK_OK) /* free it either way */
-				free(ps);
+				free(ps1);
 
 		ch = ssh_get_server_ch(pss, 0);
 
@@ -2534,6 +2511,8 @@ bail:
 		break;
 
 	case LWS_CALLBACK_CGI:
+		if (!pss)
+			break;
 		if (pss->vhd && pss->vhd->ops &&
 		    pss->vhd->ops->child_process_io &&
 		    pss->vhd->ops->child_process_io(pss->ch_temp->priv,
@@ -2542,15 +2521,19 @@ bail:
 		break;
 
 	case LWS_CALLBACK_CGI_PROCESS_ATTACH:
+		if (!pss)
+			break;
 		ch = ssh_get_server_ch(pss, pss->channel_doing_spawn);
 		if (ch) {
-			ch->spawn_pid = len; /* child process PID */
+			ch->spawn_pid = (int)len; /* child process PID */
 			lwsl_notice("associated PID %d to ch %d\n", (int)len,
 				    pss->channel_doing_spawn);
 		}
 		break;
 
 	case LWS_CALLBACK_CGI_TERMINATED:
+		if (!pss)
+			break;
 		if (pss->vhd && pss->vhd->ops &&
 		    pss->vhd->ops->child_process_terminated)
 		    pss->vhd->ops->child_process_terminated(pss->ch_temp->priv,
@@ -2565,6 +2548,7 @@ bail:
 			if (ch->spawn_pid == len) {
 				lwsl_notice("starting close of ch with PID %d\n",
 					    (int)len);
+				ch->scheduled_close = 1;
 				write_task(pss, ch, SSH_WT_CH_CLOSE);
 				break;
 			}
@@ -2604,7 +2588,7 @@ init_protocol_lws_ssh_base(struct lws_context *context,
 	}
 
 	c->protocols = protocols_sshd;
-	c->count_protocols = ARRAY_SIZE(protocols_sshd);
+	c->count_protocols = LWS_ARRAY_SIZE(protocols_sshd);
 	c->extensions = NULL;
 	c->count_extensions = 0;
 
